"""Provides a collection of Discrete Fast Fourier Transforms (DFFT) and their inverses."""

import torch

from ..core.communication import MPI
from ..core.dndarray import DNDarray
from ..core.stride_tricks import sanitize_axis
from ..core.types import heat_type_is_exact, heat_type_of, canonical_heat_type, float32
from ..core.factories import array, arange
from ..core.devices import Device

from typing import Type, Union, Tuple, Iterable, Optional

__all__ = [
    "fft",
    "fft2",
    "fftfreq",
    "fftn",
    "fftshift",
    "hfft",
    "hfft2",
    "hfftn",
    "ifft",
    "ifft2",
    "ifftn",
    "ifftshift",
    "ihfft",
    "ihfft2",
    "ihfftn",
    "irfft",
    "irfft2",
    "irfftn",
    "rfft",
    "rfft2",
    "rfftfreq",
    "rfftn",
]


def __fft_op(x: DNDarray, fft_op: callable, **kwargs) -> DNDarray:
    """
    Helper function for 1-dimensional FFT.
    """
    try:
        local_x = x.larray
    except AttributeError:
        raise TypeError(f"x must be a DNDarray, is {type(x)}")
    original_split = x.split

    # sanitize kwargs
    axis = kwargs.get("axis")
    try:
        axis = sanitize_axis(x.gshape, axis)
    except ValueError as e:
        raise IndexError(e)
    if isinstance(axis, tuple) and len(axis) > 1:
        raise TypeError(f"axis must be an integer, got {axis}")
    n = kwargs.get("n", None)
    if n is None:
        n = x.shape[axis]
    norm = kwargs.get("norm", None)

    # calculate output shape:
    # if operation requires real input, output size of last transformed dimension is the Nyquist frequency
    output_shape = list(x.shape)
    real_to_generic_fft_ops = {
        torch.fft.rfft: torch.fft.fft,
        torch.fft.ihfft: torch.fft.ifft,
    }
    real_op = fft_op in real_to_generic_fft_ops
    if real_op:
        nyquist_freq = n // 2 + 1
        output_shape[axis] = nyquist_freq
    else:
        output_shape[axis] = n

    fft_along_split = original_split == axis
    # FFT along non-split axis
    if not x.is_distributed() or not fft_along_split:
        if local_x.numel() == 0:
            # empty tensor, return empty tensor with consistent shape
            local_shape = output_shape.copy()
            local_shape[original_split] = 0
            torch_result = torch.empty(
                tuple(local_shape), dtype=local_x.dtype, device=local_x.device
            )
        else:
            torch_result = fft_op(local_x, n=n, dim=axis, norm=norm)
        return DNDarray(
            torch_result,
            gshape=tuple(output_shape),
            dtype=heat_type_of(torch_result),
            split=original_split,
            device=x.device,
            comm=x.comm,
            balanced=x.balanced,
        )

    # FFT along split axis
    if original_split != 0:
        # transpose x so redistribution starts from axis 0
        transpose_axes = list(range(x.ndim))
        transpose_axes[0], transpose_axes[original_split] = (
            transpose_axes[original_split],
            transpose_axes[0],
        )
        x = x.transpose(transpose_axes)

    # transform decomposition: split axis first, then the rest

    # redistribute x
    if x.ndim > 1:
        _ = x.resplit(axis=1)
    else:
        _ = x.resplit(axis=None)

    # if operation requires real input, switch to generic transform
    if real_op:
        fft_op = real_to_generic_fft_ops[fft_op]
    # FFT along axis 0 (now non-split)
    ht_result = __fft_op(_, fft_op, n=n, axis=0, norm=norm)
    del _
    # redistribute partial result back to axis 0
    ht_result.resplit_(axis=0)
    if original_split != 0:
        # transpose x, partial_result back to original shape
        x = x.transpose(transpose_axes)
        ht_result = ht_result.transpose(transpose_axes)

    if real_op:
        # discard elements beyond Nyquist frequency on last transformed axis
        nyquist_slice = [slice(None)] * ht_result.ndim
        nyquist_slice[axis] = slice(0, nyquist_freq)
        ht_result = ht_result[(nyquist_slice)].balance_()

    return ht_result


def __fftn_op(x: DNDarray, fftn_op: callable, **kwargs) -> DNDarray:
    """
    Helper function for fftn
    """
    try:
        local_x = x.larray
    except AttributeError:
        raise TypeError("x must be a DNDarray, is {}".format(type(x)))

    original_split = x.split
    output_shape = list(x.shape)
    shift_op = fftn_op in [torch.fft.fftshift, torch.fft.ifftshift]
    inverse_real_op = fftn_op in [torch.fft.irfftn, torch.fft.irfft2]

    torch_has_ihfftn = hasattr(torch.fft, "ihfftn")

    if torch_has_ihfftn:
        real_to_generic_fftn_ops = {
            torch.fft.rfftn: torch.fft.fftn,
            torch.fft.rfft2: torch.fft.fft2,
            torch.fft.ihfftn: torch.fft.ifftn,
            torch.fft.ihfft2: torch.fft.ifft2,
        }
    else:  # pragma: no cover
        real_to_generic_fftn_ops = {
            torch.fft.rfftn: torch.fft.fftn,
            torch.fft.rfft2: torch.fft.fft2,
        }

    real_op = fftn_op in real_to_generic_fftn_ops

    # sanitize kwargs
    if shift_op:
        # only keyword argument `axes` is supported
        axes = kwargs.get("axes", None)
        if axes is None:
            axes = tuple(range(x.ndim))
        else:
            try:
                axes = sanitize_axis(x.gshape, axes)
            except ValueError as e:
                raise IndexError(e)
        torch_kwargs = {"dim": axes}
    else:
        s = kwargs.get("s", None)
        if s is not None and len(s) > x.ndim:
            raise ValueError(
                f"Input is {x.ndim}-dimensional, so s can be at most {x.ndim} elements long. Got {len(s)} elements instead."
            )
        axes = kwargs.get("axes", None)
        if axes is None:
            if s is not None:
                axes = tuple(range(x.ndim)[-len(s) :])
            else:
                axes = tuple(range(x.ndim))
        else:
            try:
                axes = sanitize_axis(x.gshape, axes)
            except ValueError as e:
                raise IndexError(e)
        if s is None:
            s = tuple(x.shape[axis] for axis in axes)
        norm = kwargs.get("norm", None)
        for i, axis in enumerate(axes):
            output_shape[axis] = s[i]
        torch_kwargs = {"s": s, "dim": axes, "norm": norm}

    repeated_axes = axes is not None and len(axes) != len(set(axes))
    if repeated_axes:
        raise NotImplementedError("Multiple transforms over the same axis not implemented yet.")

    # calculate output shape:
    # if operation requires real input, output size of last transformed dimension is the Nyquist frequency
    if real_op:
        nyquist_freq = s[-1] // 2 + 1
        output_shape[axes[-1]] = nyquist_freq

    fft_along_split = original_split in axes
    # FFT along non-split axes only
    if not x.is_distributed() or not fft_along_split:
        if local_x.numel() == 0:
            # empty tensor, return empty tensor with consistent shape and dtype
            local_shape = output_shape.copy()
            local_shape[original_split] = 0
            if inverse_real_op:
                output_dtype = local_x.real.dtype
            else:
                # local_x is empty, memory footprint not an issue
                _ = local_x * 1j
                output_dtype = _.dtype
            torch_result = torch.empty(
                tuple(local_shape), dtype=output_dtype, device=local_x.device
            )
        else:
            torch_result = fftn_op(local_x, **torch_kwargs)
        return DNDarray(
            torch_result,
            gshape=tuple(output_shape),
            dtype=heat_type_of(torch_result),
            split=original_split,
            device=x.device,
            comm=x.comm,
            balanced=x.balanced,
        )

    # FFT along split axis
    if original_split != 0:
        # transpose x so redistribution starts from axis 0
        transpose_axes = list(range(x.ndim))
        transpose_axes[0], transpose_axes[original_split] = (
            transpose_axes[original_split],
            transpose_axes[0],
        )
        x = x.transpose(transpose_axes)

    # original split is 0 and fft is along axis 0
    if x.ndim == 1:
        _ = x.resplit(axis=None)
        ht_result = __fftn_op(_, fftn_op, **kwargs).resplit_(axis=0)
        del _
        return ht_result

    # transform decomposition: split axis first, then the rest
    # redistribute x from axis 0 to 1
    _ = x.resplit(axis=1)
    # FFT along axis 0 (now non-split)
    split_index = axes.index(original_split)
    if shift_op:
        partial_ht_result = __fftn_op(_, fftn_op, axes=(0,))
    else:
        partial_s = (s[split_index],)
        partial_ht_result = __fftn_op(_, fftn_op, s=partial_s, axes=(0,), norm=norm)
    output_shape[original_split] = partial_ht_result.shape[0]
    del _
    # redistribute partial result from axis 1 to 0
    partial_ht_result.resplit_(axis=0)
    if original_split != 0:
        # transpose x, partial_ht_result back to original shape
        x = x.transpose(transpose_axes)
        partial_ht_result = partial_ht_result.transpose(transpose_axes)

    # now apply FFT along leftover (non-split) axes
    axes = list(axes)
    axes.remove(original_split)
    axes = tuple(axes)
    if shift_op:
        ht_result = __fftn_op(partial_ht_result, fftn_op, axes=axes)
    else:
        if s is not None:
            s = list(s)
            s = s[:split_index] + s[split_index + 1 :]
            s = tuple(s)
        # if fft operation requires real input, switch to generic operation for the second pass
        if real_op:
            fftn_op = real_to_generic_fftn_ops[fftn_op]
        ht_result = __fftn_op(partial_ht_result, fftn_op, s=s, axes=axes, norm=norm)
    del partial_ht_result
    return ht_result


def __fftfreq_op(fftfreq_op: callable, **kwargs) -> DNDarray:
    """
    Helper function for ``fftfreq`` and ``rfftfreq`` operations.
    """
    n = kwargs.get("n", None)
    d = kwargs.get("d", None)
    dtype = kwargs.get("dtype", None)
    split = kwargs.get("split", None)
    device = kwargs.get("device", None)
    comm = kwargs.get("comm", None)

    if not isinstance(n, int):
        raise ValueError(f"n must be an integer, is {type(n)}")
    if not isinstance(d, (int, float)):
        if isinstance(d, complex):
            # numpy supports complex d, torch doesn't
            raise NotImplementedError("Support for complex d not implemented yet.")
        raise TypeError(f"d must be a scalar, is {type(d)}")
    if dtype is not None:
        if heat_type_is_exact(dtype):
            raise TypeError(f"dtype must be a float or complex type, is {dtype}")
        # extract torch dtype from heat dtype
        try:
            torch_dtype = dtype.torch_type()
        except AttributeError:
            raise TypeError(f"dtype must be a heat dtype, is {type(dtype)}")
    else:
        torch_dtype = None

    # early out for non-distributed fftfreq
    if split is None:
        return array(fftfreq_op(n, d=d, dtype=torch_dtype), device=device, split=None, comm=comm)

    # distributed fftfreq
    if split != 0:
        raise IndexError(f"`fftfreq` returns a 1-D array, `split` must be 0 or None, is {split}")

    # calculate parameters of the global frequency spectrum
    n_is_even = n % 2 == 0
    if n_is_even:
        middle_channel = n // 2
    else:
        middle_channel = n // 2 + 1

    # allocate global fftfreq array
    # if real operation, return only positive frequencies
    freq_dtype = (
        canonical_heat_type(torch.promote_types(torch_dtype, torch.float32))
        if torch_dtype is not None
        else float32
    )
    if fftfreq_op == torch.fft.rfftfreq:
        freqs = arange(middle_channel, dtype=freq_dtype, device=device, split=split, comm=comm)
    else:
        freqs = arange(n, dtype=freq_dtype, device=device, split=split, comm=comm)
        # second half of fftfreq returns negative frequencies in inverse order
        freqs[middle_channel:] -= n

    # calculate global frequencies
    channel_width = n * d
    freqs /= channel_width
    return freqs


def __real_fft_op(x: DNDarray, fft_op: callable, **kwargs) -> DNDarray:
    """
    Helper function for real 1-D FFTs.
    """
    if x.larray.is_complex():
        raise TypeError(f"Input array must be real, is {x.dtype}.")
    return __fft_op(x, fft_op, **kwargs)


def __real_fftn_op(x: DNDarray, fftn_op: callable, **kwargs) -> DNDarray:
    """
    Helper function for real N-D FFTs.
    """
    if x.larray.is_complex():
        raise TypeError(f"Input array must be real, is {x.dtype}.")
    return __fftn_op(x, fftn_op, **kwargs)


def fft(x: DNDarray, n: int = None, axis: int = -1, norm: str = None) -> DNDarray:
    """
    Compute the one-dimensional discrete Fourier Transform over the specified axis in an M-dimensional
    array by means of the Fast Fourier Transform (FFT). By default, the last axis is transformed, while the remaining
    axes are left unchanged.

    Parameters
    ----------
    x : DNDarray
        Input array, can be complex. WARNING: If x is 1-D and distributed, the entire array is copied on each MPI process. See Notes.
    n : int, optional
        Length of the transformed axis of the output. If not given, the length is assumed to be the length of the input
        along the axis specified by `axis`. If `n` is smaller than the length of the input, the input is truncated. If `n` is
        larger, the input is padded with zeros. Default: None.
    axis : int, optional
        Axis over which to compute the FFT. If not given, the last axis is used, or the only axis if `x` has only one
        dimension. Default: -1.
    norm : str, optional
        Normalization mode: 'forward', 'backward', or 'ortho'. Indicates in what direction the forward/backward pair of transforms is normalized. Default is "backward".

    See Also
    --------
    :func:`ifft` : inverse 1-dimensional FFT
    :func:`fft2` : 2-dimensional FFT
    :func:`fftn` : N-dimensional FFT
    :func:`rfft` : 1-dimensional FFT of a real signal
    :func:`hfft` : 1-dimensional FFT of a Hermitian symmetric sequence
    :func:`fftfreq` : frequency bins for given FFT parameters
    :func:`rfftfreq` : frequency bins for real FFT

    Notes
    -----
    This function requires MPI communication if the input array is transformed along the distribution axis.
    If the input array is 1-D and distributed, this function copies the entire array on each MPI process! i.e. if the array is very large, you might run out of memory.
    Hint: if you are looping through a batch of 1-D arrays to transform them, consider stacking them into a 2-D DNDarray and transforming them in one go (see :func:`fft2`).
    """
    return __fft_op(x, torch.fft.fft, n=n, axis=axis, norm=norm)


def fft2(
    x: DNDarray, s: Tuple[int, int] = None, axes: Tuple[int, int] = (-2, -1), norm: str = None
) -> DNDarray:
    """
    Compute the 2-dimensional discrete Fourier Transform over the specified axes in an M-dimensional
    array by means of the Fast Fourier Transform (FFT). By default, the last two axes are transformed, while the
    remaining axes are left unchanged.

    Parameters
    ----------
    x : DNDarray
        Input array, can be complex
    s : Tuple[int, int], optional
        Shape of the output along the transformed axes. (default is x.shape)
    axes : Tuple[int, int], optional
        Axes over which to compute the FFT. If not given, the last `len(s)` axes are used, or all axes if `s` is also
        not specified. Repeated transforms over an axis, i.e. repeated indices in ``axes``, are not supported yet.
        (default is (-2, -1))
    norm : str, optional
        Normalization mode: 'forward', 'backward', or 'ortho'. Indicates in what direction the forward/backward pair of transforms is normalized. Default is "backward".

    See Also
    --------
    :func:`ifft2` : inverse 2-dimensional FFT
    :func:`fft` : 1-dimensional FFT
    :func:`fftn` : N-dimensional FFT
    :func:`rfft2` : 2-dimensional FFT of a real signal
    :func:`hfft2` : 2-dimensional FFT of a Hermitian symmetric sequence

    Notes
    -----
    This function requires MPI communication if the input array is distributed and the split axis is transformed.
    """
    return __fftn_op(x, torch.fft.fft2, s=s, axes=axes, norm=norm)


def fftfreq(
    n: int,
    d: Union[int, float] = 1.0,
    dtype: Optional[Type] = None,
    split: Optional[int] = None,
    device: Optional[Union[str, Device]] = None,
    comm: Optional[MPI.Comm] = None,
) -> DNDarray:
    """
    Return the Discrete Fourier Transform sample frequencies for a signal of size ``n``.

    The returned ``DNDarray`` contains the frequency bin centers in cycles per unit of the sample spacing (with zero
    at the start). For instance, if the sample spacing is in seconds, then the frequency unit is cycles/second.

    Parameters
    ----------
    n : int
        Window length.
    d : Union[int, float], optional
        Sample spacing (inverse of the sampling rate). Defaults to 1.
    dtype : Type, optional
        The desired data type of the output. Defaults to `ht.float32`.
    split : int, optional
        The axis along which to split the result. Can be None or 0, as the output is 1-dimensional. Defaults to None, i.e. non-distributed output.
    device : str or Device, optional
        The device on which to place the output. If not given, the output is placed on the current device.
    comm : MPI.Comm, optional
        The MPI communicator to use for distributing the output. If not given, the default communicator is used.

    See Also
    --------
    :func:`rfftfreq` : frequency bins for :func:`rfft`
    """
    return __fftfreq_op(
        torch.fft.fftfreq, n=n, d=d, dtype=dtype, split=split, device=device, comm=comm
    )


def fftn(
    x: DNDarray, s: Tuple[int, ...] = None, axes: Tuple[int, ...] = None, norm: str = None
) -> DNDarray:
    """
    Compute the N-dimensional discrete Fourier Transform.

    This function computes the N-dimensional discrete Fourier Transform over any number of axes in an M-dimensional
    array by means of the Fast Fourier Transform (FFT).

    Parameters
    ----------
    x : DNDarray
        Input array, can be complex
    s : Tuple[int, ...], optional
        Shape of the output along the transformed axes. (default is x.shape)
    axes : Tuple[int, ...], optional
        Axes over which to compute the FFT. If not given, the last `len(s)` axes are used, or all axes if `s` is also
        not specified. Repeated transforms over an axis, i.e. repeated indices in ``axes``, are not supported yet.
        (default is None)
    norm : str, optional
        Normalization mode: 'forward', 'backward', or 'ortho'. Indicates in what direction the forward/backward pair of transforms is normalized. Default is "backward".

    See Also
    --------
    :func:`ifftn` : inverse N-dimensional FFT
    :func:`fft` : 1-dimensional FFT
    :func:`fft2` : 2-dimensional FFT
    :func:`rfftn` : N-dimensional FFT of a real signal
    :func:`hfftn` : N-dimensional FFT of a Hermitian symmetric sequence

    Notes
    -----
    This function requires MPI communication if the input array is distributed and the split axis is transformed.
    """
    return __fftn_op(x, torch.fft.fftn, s=s, axes=axes, norm=norm)


def fftshift(x: DNDarray, axes: Optional[Union[int, Iterable[int]]] = None) -> DNDarray:
    """
    Shift the zero-frequency component to the center of the spectrum.

    This function swaps half-spaces for all axes listed (defaults to all). Note that ``y[0]`` is the Nyquist component
    only if ``len(x)`` is even.

    Parameters
    ----------
    x : DNDarray
        Input array
    axes : int or Iterable[int], optional
        Axes over which to shift. Default is None, which shifts all axes.

    See Also
    --------
    :func:`ifftshift` : The inverse of `fftshift`.

    Notes
    -----
    This function requires MPI communication if the input array is distributed and the split axis is shifted.
    """
    return __fftn_op(x, torch.fft.fftshift, axes=axes)


def hfft(x: DNDarray, n: int = None, axis: int = -1, norm: str = None) -> DNDarray:
    """
    Compute the one-dimensional discrete Fourier Transform of a Hermitian symmetric signal.

    This function computes the one-dimensional discrete Fourier Transform over the specified axis in an M-dimensional
    array by means of the Fast Fourier Transform (FFT). By default, the last axis is transformed, while the remaining
    axes are left unchanged. The input signal is assumed to be Hermitian-symmetric, i.e. `x[..., i] = x[..., -i].conj()`.

    Parameters
    ----------
    x : DNDarray
        Input array
    n : int, optional
        Length of the transformed axis of the output.
        If `n` is not None, the input array is either zero-padded or trimmed to length `n` before the transform.
        Default: `2 * (x.shape[axis] - 1)`.
    axis : int, optional
        Axis over which to compute the FFT. If not given, the last axis is used, or the only axis if x has only one
        dimension. Default: -1.
    norm : str, optional
        Normalization mode: 'forward', 'backward', or 'ortho'. Indicates in what direction the forward/backward pair of transforms is normalized. Default is "backward".

    See Also
    --------
    :func:`ihfft` : inverse 1-dimensional FFT of a Hermitian-symmetric sequence
    :func:`hfft2` : 2-dimensional FFT of a Hermitian-symmetric sequence
    :func:`hfftn` : N-dimensional FFT of a Hermitian-symmetric sequence
    :func:`fft` : 1-dimensional FFT
    :func:`rfft` : 1-dimensional FFT of a real signal

    Notes
    -----
    This function requires MPI communication if the input array is transformed along the distribution axis.
    """
    if n is None:
        n = 2 * (x.shape[axis] - 1)
    return __fft_op(x, torch.fft.hfft, n=n, axis=axis, norm=norm)


def hfft2(
    x: DNDarray, s: Tuple[int, int] = None, axes: Tuple[int, int] = (-2, -1), norm: str = None
) -> DNDarray:
    """
    Compute the 2-dimensional discrete Fourier Transform of a Hermitian symmetric signal.

    This function computes the 2-dimensional discrete Fourier Transform over the specified axes in an M-dimensional
    array by means of the Fast Fourier Transform (FFT). By default, the last two axes are transformed, while the
    remaining axes are left unchanged. The input signal is assumed to be Hermitian-symmetric, i.e. `x[..., i] = x[..., -i].conj()`.

    Parameters
    ----------
    x : DNDarray
        Input array
    s : Tuple[int, int], optional
        Shape of the signal along the transformed axes. If `s` is specified, the input array is either zero-padded or trimmed to length `s` before the transform.
        If `s` is not given, the last dimension defaults to even output: `s[-1] = 2 * (x.shape[-1] - 1)`.
    axes : Tuple[int, int], optional
        Axes over which to compute the FFT. If not given, the last two dimensions are transformed. Repeated transforms over an axis, i.e. repeated indices in ``axes``, are not supported yet. Default: (-2, -1).
    norm : str, optional
        Normalization mode: 'forward', 'backward', or 'ortho'. Indicates in what direction the forward/backward pair of transforms is normalized. Default is "backward".

    See Also
    --------
    :func:`ihfft2` : inverse 2-dimensional FFT of a Hermitian-symmetric sequence
    :func:`hfft` : 1-dimensional FFT of a Hermitian-symmetric sequence
    :func:`hfftn` : N-dimensional FFT of a Hermitian-symmetric sequence
    :func:`fft2` : 2-dimensional FFT
    :func:`rfft2` : 2-dimensional FFT of a real signal

    Notes
    -----
    This function requires MPI communication if the input array is distributed and the split axis is transformed.
    """
    if s is None:
        s = (x.shape[axes[0]], 2 * (x.shape[axes[1]] - 1))
    return __fftn_op(x, torch.fft.hfft2, s=s, axes=axes, norm=norm)


def hfftn(
    x: DNDarray, s: Tuple[int, ...] = None, axes: Tuple[int, ...] = None, norm: str = None
) -> DNDarray:
    """
    Compute the N-dimensional discrete Fourier Transform of a Hermitian symmetric signal.

    This function computes the N-dimensional discrete Fourier Transform over any number of axes in an M-dimensional
    array by means of the Fast Fourier Transform (FFT). By default, all axes are transformed.

    Parameters
    ----------
    x : DNDarray
        Input array
    s : Tuple[int, ...], optional
        Shape of the signal along the transformed axes. If `s` is specified, the input array is either zero-padded or trimmed to length `s` before the transform.
        If `s` is not given, the last dimension defaults to even output: `s[-1] = 2 * (x.shape[-1] - 1)`.
    axes : Tuple[int, ...], optional
        Axes over which to compute the FFT. If not given, all dimensions are transformed. Repeated transforms over an axis, i.e. repeated indices in ``axes``, are not supported yet. Default: None.
    norm : str, optional
        Normalization mode: 'forward', 'backward', or 'ortho'. Indicates in what direction the forward/backward pair of transforms is normalized. Default is "backward".

    See Also
    --------
    :func:`ihfftn` : inverse N-dimensional FFT of a Hermitian-symmetric sequence
    :func:`hfft` : 1-dimensional FFT of a Hermitian-symmetric sequence
    :func:`hfft2` : 2-dimensional FFT of a Hermitian-symmetric sequence
    :func:`fftn` : N-dimensional FFT
    :func:`rfftn` : N-dimensional FFT of a real signal

    Notes
    -----
    This function requires MPI communication if the input array is distributed and the split axis is transformed.
    """
    if s is None:
        if axes is None:
            s = list(x.shape[i] for i in range(x.ndim))
        else:
            s = list(x.shape[i] for i in axes)
        s[-1] = 2 * (s[-1] - 1)
        s = tuple(s)
    return __fftn_op(x, torch.fft.hfftn, s=s, axes=axes, norm=norm)


def ifft(x: DNDarray, n: int = None, axis: int = -1, norm: str = None) -> DNDarray:
    """
    Compute the one-dimensional inverse discrete Fourier Transform.

    Parameters
    ----------
    x : DNDarray
        Input array, can be complex
    n : int, optional
        Length of the transformed axis of the output. If not given, the length is taken to be the length of the input
        along the axis specified by `axis`. If `n` is smaller than the length of the input, the input is cropped. If `n` is
        larger, the input is padded with zeros. Default: None.
    axis : int, optional
        Axis over which to compute the inverse FFT. If not given, the last axis is used, or the only axis if x has only one dimension. Default: -1.
    norm : str, optional
        Normalization mode: 'forward', 'backward', or 'ortho'. Indicates in what direction the forward/backward pair of transforms is normalized. Default is "backward".

    See Also
    --------
    :func:`fft` : forward 1-dimensional FFT
    :func:`ifft2` : inverse 2-dimensional FFT
    :func:`ifftn` : inverse N-dimensional FFT
    :func:`irfft` : inverse 1-dimensional FFT of a real sequence
    :func:`ihfft` : inverse 1-dimensional FFT of a Hermitian symmetric sequence

    Notes
    -----
    This function requires MPI communication if the input array is transformed along the distribution axis.
    If the input array is 1-D and distributed, this function copies the entire array on each MPI process! i.e. if the array is very large, you might run out of memory.
    Hint: if you are looping through a batch of 1-D arrays to transform them, consider stacking them into a 2-D DNDarray and transforming them all at once (see :func:`ifft2`).
    """
    return __fft_op(x, torch.fft.ifft, n=n, axis=axis, norm=norm)


def ifft2(
    x: DNDarray, s: Tuple[int, int] = None, axes: Tuple[int, int] = (-2, -1), norm: str = None
) -> DNDarray:
    """
    Compute the 2-dimensional inverse discrete Fourier Transform.

    Parameters
    ----------
    x : DNDarray
        Input array, can be complex
    s : Tuple[int, int], optional
        Shape of the output along the transformed axes. (default is x.shape)
    axes : Tuple[int, int], optional
        Axes over which to compute the inverse FFT. If not given, the last `len(s)` axes are used, or all axes if `s` is
        also not specified. Repeated transforms over an axis, i.e. repeated indices in ``axes``, are not supported yet. Default: (-2, -1).
    norm : str, optional
        Normalization mode: 'forward', 'backward', or 'ortho'. Indicates in what direction the forward/backward pair of transforms is normalized. Default is "backward".

    See Also
    --------
    :func:`fft2` : forward 2-dimensional FFT
    :func:`ifft` : inverse 1-dimensional FFT
    :func:`ifftn` : inverse N-dimensional FFT
    :func:`irfft2` : inverse 2-dimensional FFT of a real sequence
    :func:`ihfft2` : inverse 2-dimensional FFT of a Hermitian symmetric sequence

    Notes
    -----
    This function requires MPI communication if the input array is distributed and the split axis is transformed.
    """
    return __fftn_op(x, torch.fft.ifft2, s=s, axes=axes, norm=norm)


def ifftn(
    x: DNDarray, s: Tuple[int, int] = None, axes: Tuple[int, ...] = None, norm: str = None
) -> DNDarray:
    """
    Compute the N-dimensional inverse discrete Fourier Transform.

    Parameters
    ----------
    x : DNDarray
        Input array, can be complex
    s : Tuple[int, ...], optional
        Shape of the output along the transformed axes. (default is x.shape)
    axes : Tuple[int, ...], optional
        Axes over which to compute the inverse FFT. If not given, the last `len(s)` axes are used, or all axes if `s` is
        also not specified. Repeated transforms over an axis, i.e. repeated indices in ``axes``, are not supported yet. Default: None.
    norm : str, optional
        Normalization mode: 'forward', 'backward', or 'ortho'. Indicates in what direction the forward/backward pair of transforms is normalized. Default is "backward".

    See Also
    --------
    :func:`fftn` : forward N-dimensional FFT
    :func:`ifft` : inverse 1-dimensional FFT
    :func:`ifft2` : inverse 2-dimensional FFT
    :func:`irfftn` : inverse N-dimensional FFT of a real sequence
    :func:`ihfftn` : inverse N-dimensional FFT of a Hermitian symmetric sequence

    Notes
    -----
    This function requires MPI communication if the input array is distributed and the split axis is transformed.
    """
    return __fftn_op(x, torch.fft.ifftn, s=s, axes=axes, norm=norm)


def ifftshift(x: DNDarray, axes: Optional[Union[int, Iterable[int]]] = None) -> DNDarray:
    """
    The inverse of fftshift.

    Parameters
    ----------
    x : DNDarray
        Input array
    axes : int or Iterable[int], optional
        Axes over which to shift. Default is None, which shifts all axes.

    See Also
    --------
    :func:`fftshift` : Shift the zero-frequency component to the center of the spectrum.

    Notes
    -----
    This function requires MPI communication if the input array is distributed and the split axis is shifted.
    """
    return __fftn_op(x, torch.fft.ifftshift, axes=axes)


def ihfft(x: DNDarray, n: int = None, axis: int = -1, norm: str = None) -> DNDarray:
    """
    Compute the one-dimensional inverse discrete Fourier Transform of a real signal. The output is Hermitian-symmetric.

    Parameters
    ----------
    x : DNDarray
        Input array, must be real
    n : int, optional
        Length of the transformed axis of the output. If not given, the length is taken to be the length of the input
        along the axis specified by `axis`. If `n` is smaller than the length of the input, the input is cropped. If `n` is
        larger, the input is padded with zeros. Default: None.
    axis : int, optional
        Axis over which to compute the inverse FFT. If not given, the last axis is used, or the only axis if x has only one dimension. Default: -1.
    norm : str, optional
        Normalization mode: 'forward', 'backward', or 'ortho'. Indicates in what direction the forward/backward pair of transforms is normalized. Default is "backward".

    See Also
    --------
    :func:`hfft` : 1-dimensional FFT of a Hermitian-symmetric sequence
    :func:`ihfft2` : inverse 2-dimensional FFT of a Hermitian-symmetric sequence
    :func:`ihfftn` : inverse N-dimensional FFT of a Hermitian-symmetric sequence
    :func:`rfft` : 1-dimensional FFT of a real signal
    :func:`irfft` : inverse 1-dimensional FFT of a real sequence

    Notes
    -----
    This function requires MPI communication if the input array is transformed along the distribution axis.
    """
    return __real_fft_op(x, torch.fft.ihfft, n=n, axis=axis, norm=norm)


def ihfft2(
    x: DNDarray, s: Tuple[int, int] = None, axes: Tuple[int, int] = (-2, -1), norm: str = None
) -> DNDarray:
    """
    Compute the inverse of a 2-dimensional discrete Fourier Transform of a Hermitian-symmetric signal. The output is Hermitian-symmetric. Requires torch >= 1.11.0.

    Parameters
    ----------
    x : DNDarray
        Input array, must be real
    s : Tuple[int, int], optional
        Shape of the output along the transformed axes. (default is x.shape)
    axes : Tuple[int, int], optional
        Axes over which to compute the inverse FFT. If not given, the last `len(s)` axes are used, or all axes if `s` is
        also not specified. Repeated transforms over an axis, i.e. repeated indices in ``axes``, are not supported yet. Default is (-2, -1).
    norm : str, optional
        Normalization mode: 'forward', 'backward', or 'ortho'. Indicates in what direction the forward/backward pair of transforms is normalized. Default is "backward".

    See Also
    --------
    :func:`hfft2` : 2-dimensional FFT of a Hermitian-symmetric sequence
    :func:`ihfft` : inverse 1-dimensional FFT of a Hermitian-symmetric sequence
    :func:`ihfftn` : inverse N-dimensional FFT of a Hermitian-symmetric sequence
    :func:`rfft2` : 2-dimensional FFT of a real signal
    :func:`irfft2` : inverse 2-dimensional FFT of a real sequence

    Notes
    -----
    This function requires MPI communication if the input array is distributed and the split axis is transformed.
    """
    torch_has_ihfftn = hasattr(torch.fft, "ihfftn")
    if not torch_has_ihfftn:  # pragma: no cover
        raise NotImplementedError(
            f"n-dim inverse Hermitian FFTs not implemented for torch < 1.11.0. Your environment runs torch {torch.__version__}. Please upgrade torch."
        )
    return __real_fftn_op(x, torch.fft.ihfft2, s=s, axes=axes, norm=norm)


def ihfftn(
    x: DNDarray, s: Tuple[int, ...] = None, axes: Tuple[int, ...] = None, norm: str = None
) -> DNDarray:
    """
    Compute the inverse of a N-dimensional discrete Fourier Transform of Hermitian-symmetric signal. The output is Hermitian-symmetric. Requires torch >= 1.11.0.

    Parameters
    ----------
    x : DNDarray
        Input array, must be real
    s : Tuple[int, ...], optional
        Shape of the output along the transformed axes. (default is x.shape)
    axes : Tuple[int, ...], optional
        Axes over which to compute the inverse FFT. If not given, the last `len(s)` axes are used, or all axes if `s` is
        also not specified. Repeated transforms over an axis, i.e. repeated indices in ``axes``, are not supported yet. Default: None.
    norm : str, optional
        Normalization mode: 'forward', 'backward', or 'ortho'. Indicates in what direction the forward/backward pair of transforms is normalized. Default is "backward".

    See Also
    --------
    :func:`hfftn` : N-dimensional FFT of a Hermitian-symmetric sequence
    :func:`ihfft` : inverse 1-dimensional FFT of a Hermitian-symmetric sequence
    :func:`ihfft2` : inverse 2-dimensional FFT of a Hermitian-symmetric sequence
    :func:`rfftn` : N-dimensional FFT of a real signal
    :func:`irfftn` : inverse N-dimensional FFT of a real sequence

    Notes
    -----
    This function requires MPI communication if the input array is distributed and the split axis is transformed.
    """
    torch_has_ihfftn = hasattr(torch.fft, "ihfftn")
    if not torch_has_ihfftn:  # pragma: no cover
        raise NotImplementedError(
            f"n-dim inverse Hermitian FFTs not implemented for torch < 1.11.0. Your environment runs torch {torch.__version__}. Please upgrade torch."
        )
    return __real_fftn_op(x, torch.fft.ihfftn, s=s, axes=axes, norm=norm)


def irfft(x: DNDarray, n: int = None, axis: int = -1, norm: str = None) -> DNDarray:
    """
    Compute the inverse of a one-dimensional discrete Fourier Transform of real signal. The output is real.

    Parameters
    ----------
    x : DNDarray
        Input array, can be complex
    n : int, optional
        Length of the transformed axis of the output. If not given, the length is taken to be the length of the input
        along the axis specified by `axis`. If `n` is smaller than the length of the input, the input is cropped. If `n` is
        larger, the input is padded with zeros. Default: None.
    axis : int, optional
        Axis over which to compute the inverse FFT. If not given, the last axis is used, or the only axis if x has only one dimension. Default: -1.
    norm : str, optional
        Normalization mode: 'forward', 'backward', or 'ortho'. Indicates in what direction the forward/backward pair of transforms is normalized. Default is "backward".

    See Also
    --------
    :func:`irfft2` : inverse 2-dimensional FFT
    :func:`irfftn` : inverse N-dimensional FFT
    :func:`rfft` : 1-dimensional FFT of a real signal
    :func:`hfft` : 1-dimensional FFT of a Hermitian symmetric sequence
    :func:`fft` : 1-dimensional FFT

    Notes
    -----
    This function requires MPI communication if the input array is transformed along the distribution axis.
    If the input array is 1-D and distributed, this function copies the entire array on each MPI process! i.e. if the array is very large, you might run out of memory.
    Hint: if you are looping through a batch of 1-D arrays to transform them, consider stacking them into a 2-D DNDarray and transforming them all at once (see :func:`irfft2`).
    """
    if n is None:
        n = 2 * (x.shape[axis] - 1)
    return __fft_op(x, torch.fft.irfft, n=n, axis=axis, norm=norm)


def irfft2(
    x: DNDarray, s: Tuple[int, int] = None, axes: Tuple[int, int] = (-2, -1), norm: str = None
) -> DNDarray:
    """
    Compute the inverse of a 2-dimensional discrete real Fourier Transform. The output is real.

    Parameters
    ----------
    x : DNDarray
        Input array, can be complex
    s : Tuple[int, int], optional
        Shape of the output along the transformed axes.
    axes : Tuple[int, int], optional
        Axes over which to compute the inverse FFT. If not given, the last `len(s)` axes are used, or all axes if `s` is
        also not specified. Repeated transforms over an axis, i.e. repeated indices in ``axes``, are not supported yet. Default is (-2, -1))
    norm : str, optional
        Normalization mode: 'forward', 'backward', or 'ortho'. Indicates in what direction the forward/backward pair of transforms is normalized. Default is "backward".

    See Also
    --------
    :func:`irfft` : inverse 1-dimensional FFT
    :func:`irfftn` : inverse N-dimensional FFT
    :func:`rfft2` : 2-dimensional FFT of a real signal
    :func:`hfft2` : 2-dimensional FFT of a Hermitian symmetric sequence
    :func:`fft2` : 2-dimensional FFT

    Notes
    -----
    This function requires MPI communication if the input array is distributed and the split axis is transformed.
    """
    if s is None:
        s = (x.shape[axes[0]], 2 * (x.shape[axes[1]] - 1))
    return __fftn_op(x, torch.fft.irfft2, s=s, axes=axes, norm=norm)


def irfftn(
    x: DNDarray, s: Tuple[int, int] = None, axes: Tuple[int, ...] = None, norm: str = None
) -> DNDarray:
    """
    Compute the inverse of an N-dimensional discrete Fourier Transform of real signal.
    The output is real.

    Parameters
    ----------
    x : DNDarray
        Input array, assumed to be Hermitian-symmetric along the transformed axes, with the last transformed axis only containing the positive half of the frequencies.
    s : Tuple[int, ...], optional
        Shape of the output along the transformed axes. If ``s`` is not specified, the last transposed axis is reconstructued in full, i.e. `s[-1] = 2 * (x.shape[axes[-1]] - 1)`.
    axes : Tuple[int, ...], optional
        Axes over which to compute the inverse FFT. If not given, the last `len(s)` axes are used, or all axes if `s` is
        also not specified. Repeated transforms over an axis, i.e. repeated indices in ``axes``, are not supported yet.
        (default is None)
    norm : str, optional
        Normalization mode: 'forward', 'backward', or 'ortho'. Indicates in what direction the forward/backward pair of transforms is normalized. Default is "backward".

    Notes
    -----
    This function requires MPI communication if the input array is distributed and the split axis is transformed.
    """
    if s is None:
        if axes is None:
            s = list(x.shape[i] for i in range(x.ndim))
        else:
            s = list(x.shape[i] for i in axes)
        s[-1] = 2 * (s[-1] - 1)
        s = tuple(s)
    return __fftn_op(x, torch.fft.irfftn, s=s, axes=axes, norm=norm)


def rfft(x: DNDarray, n: int = None, axis: int = -1, norm: str = None) -> DNDarray:
    """
    Compute the one-dimensional discrete Fourier Transform of real input. The output is Hermitian-symmetric.

    Parameters
    ----------
    x : DNDarray
        Input array, must be real.
    n : int, optional
        Length of the transformed axis of the output. If not given, the length is taken to be the length of the input
        along the axis specified by `axis`. If `n` is smaller than the length of the input, the input is cropped. If `n` is
        larger, the input is padded with zeros. Default: None.
    axis : int, optional
        Axis over which to compute the FFT. If not given, the last axis is used, or the only axis if x has only one dimension. Default: -1.
    norm : str, optional
        Normalization mode: 'forward', 'backward', or 'ortho'. Indicates in what direction the forward/backward pair of transforms is normalized. Default is "backward".

    Notes
    -----
    This function requires MPI communication if the input array is transformed along the distribution axis.
    If the input array is 1-D and distributed, this function copies the entire array on each MPI process! i.e. if the array is very large, you might run out of memory.
    Hint: if you are looping through a batch of 1-D arrays to transform them, consider stacking them into a 2-D DNDarray and transforming them all at once (see :func:`rfft2`).
    """
    return __real_fft_op(x, torch.fft.rfft, n=n, axis=axis, norm=norm)


def rfft2(
    x: DNDarray, s: Tuple[int, int] = None, axes: Tuple[int, int] = (-2, -1), norm: str = None
) -> DNDarray:
    """
    Compute the 2-dimensional discrete Fourier Transform of real input. The output is Hermitian-symmetric.

    Parameters
    ----------
    x : DNDarray
        Input array, must be real.
    s : Tuple[int, int], optional
        Shape of the output along the transformed axes. (default is x.shape)
    axes : Tuple[int, int], optional
        Axes over which to compute the FFT. If not given, the last `len(s)` axes are used, or all axes if `s` is
        also not specified. Repeated transforms over an axis, i.e. repeated indices in ``axes``, are not supported yet. (default is (-2, -1))
    norm : str, optional
        Normalization mode: 'forward', 'backward', or 'ortho'. Indicates in what direction the forward/backward pair of transforms is normalized. Default is "backward".

    Notes
    -----
    This function requires MPI communication if the input array is distributed and the split axis is transformed.
    """
    return __real_fftn_op(x, torch.fft.rfft2, s=s, axes=axes, norm=norm)


def rfftfreq(
    n: int,
    d: Union[int, float] = 1.0,
    dtype: Optional[Type] = None,
    split: Optional[int] = None,
    device: Optional[Union[str, Device]] = None,
    comm: Optional[MPI.Comm] = None,
) -> DNDarray:
    """
    Return the Discrete Fourier Transform sample frequencies.

    The returned float DNDarray contains the frequency bin centers in cycles per unit of the sample spacing (with zero
    at the start). For instance, if the sample spacing is in seconds, then the frequency unit is cycles/second.

    Parameters
    ----------
    n : int
        Window length.
    d : Union[int, float], optional
        Sample spacing (inverse of the sampling rate). Defaults to 1.
    dtype : Type, optional
        The desired data type of the output. Defaults to `float32`.
    split : int, optional
        The axis along which to split the result. If not given, the result is not split.
    device : str or Device, optional
        The device on which to place the output. If not given, the output is placed on the current device.
    comm : MPI.Comm, optional
        The MPI communicator to use for distributing the output. If not given, the default communicator is used.
    """
    return __fftfreq_op(
        torch.fft.rfftfreq, n=n, d=d, dtype=dtype, split=split, device=device, comm=comm
    )


def rfftn(
    x: DNDarray, s: Tuple[int, int] = None, axes: Tuple[int, ...] = None, norm: str = None
) -> DNDarray:
    """
    Compute the N-dimensional discrete Fourier Transform of real input. By default, all axes are transformed, with the real transform
    performed over the last axis, while the remaining transforms are complex. The output is Hermitian-symmetric, with the last transformed axis having length `s[-1] // 2 + 1` (the positive part of the spectrum).

    Parameters
    ----------
    x : DNDarray
        Input array, must be real.
    s : Tuple[int, ...], optional
        Shape of the output along the transformed axes.
    axes : Tuple[int, ...], optional
        Axes over which to compute the FFT. If not given, the last `len(s)` axes are used, or all axes if `s` is
        also not specified. Repeated transforms over an axis, i.e. repeated indices in ``axes``, are not supported yet. (default is None)
    norm : str, optional
        Normalization mode: 'forward', 'backward', or 'ortho'. Indicates in what direction the forward/backward pair of transforms is normalized. Default is "backward".

    Notes
    -----
    This function requires MPI communication if the input array is distributed and the split axis is transformed.
    """
    return __real_fftn_op(x, torch.fft.rfftn, s=s, axes=axes, norm=norm)



"""
Module implementing basic data preprocessing techniques
"""

import heat as ht
from typing import Optional, Tuple, Union

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

"""
The implementation is heavily inspired by the corresponding routines in scikit-learn (https://scikit-learn.org/stable/modules/preprocessing.html).
"""


# auxiliary function that checks if input array is appropriate to serve as data set for fitting or transforming
def _is_2D_float_DNDarray(input):
    if not isinstance(input, ht.DNDarray):
        raise TypeError(f"Input of preprocessing routines must be DNDarray, but is {type(input)}.")
    if not input.ndim == 2:
        raise ValueError(
            f"Input of preprocessing routines must be a 2D DNDarray of shape (n_datapoints, n_features), but dimension is {input.ndim}."
        )
    if ht.heat_type_is_exact(input.dtype):
        raise TypeError(
            f"Supported data types for preprocessing routines are float32 and float64, but dtype of input is {input.dtype}.",
        )


# auxiliary function that checks whether parameter of a Scaler and data to be transformed with this Scaler have matching shapes
def _has_n_features(param, inputdata):
    if param.shape[0] != inputdata.shape[1]:
        raise ValueError(
            f"Scaler has been fitted on a data set with {param.shape[0]} features, but shall now be applied to data with {inputdata.shape[1]} features."
        )


# auxiliary function that returns expected precision depending on input data type
# this is used to determine whether a feature is almost constant (w.r.t. machine precision) and should therefore not be scaled
def _tol_wrt_dtype(inputdata):
    if inputdata.dtype == ht.float32:
        return 1e-7
    if inputdata.dtype == ht.float64:
        return 1e-14


class StandardScaler(ht.TransformMixin, ht.BaseEstimator):
    """Standardization of features to mean 0 and variance 1 by affine linear transformation; similar to `sklearn.preprocessing.StandardScaler`.
    The data set to be scaled must be stored as 2D-`DNDarray` of shape (n_datapoints, n_features).
    Shifting to mean 0 and scaling to variance 1 is applied to each feature independently.

    Parameters
    ----------
    copy : bool, default=True
        If False, try to avoid a copy and do inplace scaling instead.

    with_mean : bool, default=True
        If True, center the data (i.e. mean = 0) before scaling.

    with_std : bool, default=True
        If True, scale the data to variance = 1.

    Attributes
    ----------
    scale_ : DNDarray of shape (n_features,) or None
        Per feature relative scaling of the data to achieve unit
        variance. Set to ``None`` (no variance scaling applied) if ``var = None`` or ``var`` below machine precision.

    mean_ : DNDarray of shape (n_features,) or None
        The mean value for each feature. Equal to ``None`` when ``with_mean=False``.

    var_ : DNDarray of shape (n_features,) or None
        Featurewise variance of the given data. Equal to ``None`` when ``with_std=False``.
    """

    def __init__(self, *, copy: bool = True, with_mean: bool = True, with_std: bool = True):
        self.with_mean = with_mean
        self.with_std = with_std
        self.copy = copy

    def fit(self, X: ht.DNDarray, sample_weight: Optional[ht.DNDarray] = None) -> Self:
        """
        Fit ``StandardScaler`` to the given data ``X``, i.e. compute mean and standard deviation of ``X`` to be used for later scaling.

        Parameters
        ----------
        X : DNDarray of shape (n_datapoints, n_features).
            Data used to compute the mean and standard deviation used for later featurewise scaling.

        sample_weight : Not yet supported.
            Raises ``NotImplementedError``.
        """
        if sample_weight is not None:
            NotImplementedError(
                "Standard scaler with sample weights is not yet implemented. You can open an issue to request this feature on  https://github.com/helmholtz-analytics/heat/issues/new/choose."
            )
        _is_2D_float_DNDarray(X)

        # determine mean and variance of the input data X and store them in self.mean_ and self.var_
        self.mean_ = ht.mean(X, axis=0)
        self.var_ = ht.var(X, axis=0)

        # check if var_ is below machine precision for some features, set scaling factor to 1 for these features if so and print warning
        self.scale_ = self.var_
        tol = _tol_wrt_dtype(X)
        if self.scale_.min() < tol:
            self.scale_ = ht.where(
                ht.abs(self.scale_) >= tol, self.scale_, ht.ones_like(self.scale_)
            )
            print(
                "At least one of the features is almost constant (w.r.t. machine precision) and will not be scaled for this reason."
            )
        self.scale_ = 1.0 / (self.scale_) ** 0.5
        return self

    def transform(self, X: ht.DNDarray) -> Union[Self, ht.DNDarray]:
        """Applies standardization to input data ``X`` by centering and scaling w.r.t. mean and std previously computed and saved in ``StandardScaler`` with :meth:``fit``.

        Parameters
        ----------
        X : DNDarray (n_datapoints, n_features)
            The data set to be standardized.
        copy : bool, default=None
            Copy the input ``X`` or not.
        """
        _is_2D_float_DNDarray(X)
        _has_n_features(self.mean_, X)
        if self.copy:
            return (X - self.mean_) * self.scale_
        # else in-place:
        X -= self.mean_
        X *= self.scale_
        return self

    def inverse_transform(self, Y: ht.DNDarray) -> Union[Self, ht.DNDarray]:
        """
        Scale back the data to the original representation, i.e. apply the inverse of :meth:``transform`` to the input ``Y``.

        Parameters
        ----------
        Y : DNDarray of shape (n_datapoints, n_features)
            Data to be scaled back.
        copy : bool, default=None
            Copy the input ``Y`` or not.
        """
        _is_2D_float_DNDarray(Y)
        _has_n_features(self.mean_, Y)
        if self.copy:
            return Y / self.scale_ + self.mean_
        # else in-place:
        Y /= self.scale_
        Y += self.mean_
        return self


class MinMaxScaler(ht.TransformMixin, ht.BaseEstimator):
    """
    Min-Max-Scaler: transforms the features by scaling each feature (affine) linearly to the prescribed range;
    similar to `sklearn.preprocessing.MinMaxScaler`.
    The data set to be scaled must be stored as 2D-`DNDarray` of shape (n_datapoints, n_features).

    Each feature is scaled and translated individually such that it is in the given range on the input data set,
    e.g. between zero and one (default).

    Parameters
    ----------
    feature_range : tuple (min, max), default=(0, 1)
        Desired range of transformed features.

    copy : bool, default=True
        ``copy = False`` means in-place transformations whenever possible.

    clip : Not yet supported.
        raises ``NotImplementedError``.

    Attributes
    ----------
    min_ : DNDarray of shape (n_features,)
        translation required per feature

    scale_ : DNDarray of shape (n_features,)
        scaling required per feature

    data_min_ : DNDarray of shape (n_features,)
        minimum per feature in the input data set

    data_max_ : DNDarray of shape (n_features,)
        maximum per feature in the input data set

    data_range_ : DNDarray of shape (n_features,)
        range per feature in the input data set
    """

    def __init__(
        self,
        feature_range: Tuple[float, float] = (0.0, 1.0),
        *,
        copy: bool = True,
        clip: bool = False,
    ):
        self.copy = copy
        self.feature_range = feature_range
        self.clip = clip
        if clip:
            raise NotImplementedError(
                "Clipped MinMaxScaler is not yet implemented. You can open an issue to request this feature on  https://github.com/helmholtz-analytics/heat/issues/new/choose."
            )
        if feature_range[1] <= feature_range[0]:
            raise ValueError(
                f"Upper bound of feature_range must be strictly larger than lower bound, but provided bounds are {self.feature_range[0]} and {self.feature_range[1]}."
            )

    def fit(self, X: ht.DNDarray) -> Self:
        """
        Fit the MinMaxScaler: i.e. compute the parameters required for later scaling.

        Parameters
        ----------
        X : DNDarray of shape (n_datapoints, n_features)
            data set to which scaler shall be fitted.
        """
        _is_2D_float_DNDarray(X)
        self.data_min_ = ht.min(X, axis=0)
        self.data_max_ = ht.max(X, axis=0)
        self.data_range_ = self.data_max_ - self.data_min_

        # if data_range is below machine precision for a feature, set scaling factor to 1 for this feature
        tol = _tol_wrt_dtype(X)
        self.scale_ = self.data_range_ / (self.feature_range[1] - self.feature_range[0])
        if ht.abs(self.data_range_).min() < tol:
            self.scale_ = ht.where(
                ht.abs(self.data_range_) >= tol, self.scale_, ht.ones_like(self.data_range_)
            )
            print(
                "At least one of the features is almost constant (w.r.t. machine precision) and will not be scaled for this reason."
            )
        self.scale_ = 1.0 / self.scale_
        self.min_ = -self.data_min_ * self.scale_ + self.feature_range[0]
        return self

    def transform(self, X: ht.DNDarray) -> Union[Self, ht.DNDarray]:
        """
        Transform input data with MinMaxScaler: i.e. scale features of ``X`` according to feature_range.

        Parameters
        ----------
        X : DNDarray of shape (n_datapoints, n_features)
            Data set to be transformed.
        """
        _is_2D_float_DNDarray(X)
        _has_n_features(self.data_min_, X)
        if self.copy:
            Y = (X - self.data_min_) * self.scale_ + self.feature_range[0]
            return Y
        # else in-place:
        X -= self.data_min_
        X *= self.scale_
        X += self.feature_range[0]
        return self

    def inverse_transform(self, Y: ht.DNDarray) -> Union[Self, ht.DNDarray]:
        """
        Apply the inverse of :meth:``fit``.

        Parameters
        ----------
        Y : DNDarray of shape (n_datapoints, n_features)
            Data set to be transformed back.
        """
        _is_2D_float_DNDarray(Y)
        _has_n_features(self.data_min_, Y)
        if self.copy:
            X = (Y - self.feature_range[0]) / self.scale_ + self.data_min_
            return X
        # else in-place:
        Y -= self.feature_range[0]
        Y /= self.scale_
        Y += self.data_min_
        return self


class Normalizer(ht.TransformMixin, ht.BaseEstimator):
    """
    Normalizer: each data point of a data set is scaled to unit norm independently.
    The data set to be scaled must be stored as 2D-`DNDarray` of shape (n_datapoints, n_features); therefore
    the Normalizer scales each row to unit norm. This object is similar to `sklearn.preprocessing.Normalizer`.

    Parameters
    ----------
    norm : {'l1', 'l2', 'max'}, default='l2'
        The norm to use to normalize the data points. ``norm='max'`` refers to the :math:`\\ell^\\infty`-norm.

    copy : bool, default=True
        ``copy=False`` enables in-place normalization.

    Attributes
    ----------
    None


    Notes
    -----
    Normalizer is :term:`stateless` and, consequently, :meth:``fit`` is only a dummy that does not need to be called before :meth:``transform``.
    Since :meth:``transform`` is not bijective, there is no back-transformation :meth:``inverse_transform``.
    """

    def __init__(self, norm: str = "l2", *, copy: bool = True):
        self.norm_ = norm
        self.copy = copy
        if norm == "l2":
            self.ord_ = 2
        elif norm == "l1":
            self.ord_ = 1
        elif norm == "max":
            self.ord_ = ht.inf
        else:
            raise NotImplementedError(
                "Normalization with respect to norms other than l2, l1 or linfty not yet implemented. You can open an issue to request this feature on  https://github.com/helmholtz-analytics/heat/issues/new/choose."
            )

    def fit(self, X: ht.DNDarray) -> Self:
        """Since :object:``Normalizer`` is stateless, this function is only a dummy."""
        return self

    def transform(self, X: ht.DNDarray) -> Union[Self, ht.DNDarray]:
        """
        Apply Normalizer trasformation: scales each data point of the input data set ``X`` to unit norm (w.r.t. to ``norm``).

        Parameters
        ----------
        X : DNDarray of shape (n_datapoints, n_features)
            The data set to be normalized.

        copy : bool, default=None
            ``copy=False`` enables in-place transformation.
        """
        _is_2D_float_DNDarray(X)
        X_norms = ht.norm(X, axis=1, ord=self.ord_).reshape((-1, 1))

        # if norm of data point is close to zero (w.r.t. machine precision), do not scale this data point
        tol = _tol_wrt_dtype(X)
        if X_norms.min() < tol:
            X_norms = ht.where(X_norms >= tol, X_norms, ht.ones_like(X_norms))
            print(
                "At least one of the data points has almost zero norm (w.r.t. machine precision) and will not be scaled for this reason."
            )
        if self.copy:
            Y = X / X_norms
            return Y
        # else in-place:
        X /= X_norms
        del X_norms
        return self


class MaxAbsScaler(ht.TransformMixin, ht.BaseEstimator):
    """
    MaxAbsScaler: scale each feature of a given data set linearly by its maximum absolute value. The underyling data set to be scaled is
    assumed to be stored as a 2D-`DNDarray` of shape (n_datapoints, n_features); this routine is similar to
    `sklearn.preprocessing.MaxAbsScaler`.

    Each feature is scaled individually such that the maximal absolute value of each feature after transformation will be 1.0.
    No shifting/centering is applied.

    Parameters
    ----------
    copy : bool, default=True
        ``copy=False`` enables in-place transformation.

    Attributes
    ----------
    scale_ : DNDarray of shape (n_features,)
        Per feature relative scaling of the data.

    max_abs_ : DNDarray of shape (n_features,)
        Per feature maximum absolute value of the input data.
    """

    def __init__(self, *, copy: bool = True):
        self.copy = copy

    def fit(self, X: ht.DNDarray) -> Self:
        """
        Fit MaxAbsScaler to input data ``X``: compute the parameters to be used for later scaling.

        Parameters
        ----------
        X : DNDarray of shape (n_datapoints, n_features)
            The data set to which the scaler shall be fitted.
        """
        _is_2D_float_DNDarray(X)
        self.max_abs_ = ht.norm(X, axis=0, ord=ht.inf)

        # if max abs is close to machine precision for some feature, do not scale this feature
        tol = _tol_wrt_dtype(X)
        self.scale_ = self.max_abs_
        if self.scale_.min() < tol:
            self.scale_ = ht.where(self.scale_ >= tol, self.scale_, ht.ones_like(self.scale_))
            print(
                "At least one of the features is almost constant zero (w.r.t. machine precision) and will not be scaled for this reason."
            )
        self.scale_ = 1.0 / self.scale_
        return self

    def transform(self, X: ht.DNDarray) -> Union[Self, ht.DNDarray]:
        """
        Scale the data with the MaxAbsScaler.

        Parameters
        ----------
        X : DNDarray of shape (n_datapoints, n_features)
            The data set to be scaled.
        """
        _is_2D_float_DNDarray(X)
        _has_n_features(self.scale_, X)
        if self.copy:
            Y = X * self.scale_
            return Y
        # else in-place:
        X *= self.scale_
        return self

    def inverse_transform(self, Y: ht.DNDarray) -> Union[Self, ht.DNDarray]:
        """
        Apply the inverse of :meth:``transform``, i.e. scale the input data ``Y`` back to the original representation.

        Parameters
        ----------
        Y : DNDarray of shape (n_datapoints, n_features)
            The data set to be transformed back.
        """
        _is_2D_float_DNDarray(Y)
        _has_n_features(self.scale_, Y)
        if self.copy:
            X = Y / self.scale_
            return X
        # else in-place:
        Y /= self.scale_
        return self


class RobustScaler(ht.TransformMixin, ht.BaseEstimator):
    """
    This scaler transforms the features of a given data set making use of statistics
    that are robust to outliers: it removes the median and scales the data according to
    the quantile range (defaults to IQR: Interquartile Range); this routine is similar
    to ``sklearn.preprocessing.RobustScaler``.

    The underyling data set to be scaled must be stored as a 2D-`DNDarray` of shape (n_datapoints, n_features).
    Each feature is centered and scaled independently.

    Parameters
    ----------
    with_centering : bool, default=True
        If `True`, data are centered before scaling.

    with_scaling : bool, default=True
        If `True`, scale the data to prescribed interquantile range.

    quantile_range : tuple (q_min, q_max), 0.0 <= q_min < q_max <= 100.0, \
        default=(25.0, 75.0)
        Quantile range used to calculate `scale_`; default is the so-called
        the IQR given by ``q_min=25`` and ``q_max=75``.

    copy : bool, default=True
        ``copy=False`` enable in-place transformations.

    unit_variance : not yet supported.
        raises ``NotImplementedError``

    Attributes
    ----------
    center_ : DNDarray of shape (n_features,)
        Feature-wise median value of the given data set.

    iqr_ : DNDarray of shape (n_features,)
        length of the interquantile range for each feature.

    scale_ : array of floats
        feature-wise inverse of ``iqr_``.
    """

    def __init__(
        self,
        *,
        with_centering: bool = True,
        with_scaling: bool = True,
        quantile_range: Tuple[float, float] = (25.0, 75.0),
        copy: bool = True,
        unit_variance: bool = False,
    ):
        self.with_centering = with_centering
        self.with_scaling = with_scaling
        self.quantile_range = quantile_range
        self.copy = copy
        if not with_centering and not with_scaling:
            raise ValueError(
                "Both centering and scaling are disabled, thus RobustScaler could do nothing. At least one of with_scaling or with_centering must be True."
            )
        if (
            self.quantile_range[0] >= self.quantile_range[1]
            or self.quantile_range[1] > 100.0
            or self.quantile_range[0] < 0.0
        ):
            raise ValueError(
                f"Lower bound of quantile range must be strictly smaller than uppert bound and both bounds need to be between 0.0 and 100.0. Inputs however are {self.quantile_range[0]} and {self.quantile_range[1]}."
            )
        if unit_variance:
            raise NotImplementedError(
                "Robust Scaler with additional unit variance scaling is not yet implemented. You can open an issue to request this feature on  https://github.com/helmholtz-analytics/heat/issues/new/choose."
            )
        else:
            self.unit_variance = unit_variance

    def fit(self, X: ht.DNDarray) -> Self:
        """
        Fit RobustScaler to given data set, i.e. compute the parameters required for transformation.

        Parameters
        ----------
        X : DNDarray of shape (n_datapoints, n_features)
            Data to which the Scaler should be fitted.
        """
        _is_2D_float_DNDarray(X)
        if self.with_centering:
            self.center_ = ht.median(X, axis=0)
        if self.with_scaling:
            self.iqr_ = ht.percentile(X, self.quantile_range[1], axis=0) - ht.percentile(
                X, self.quantile_range[0], axis=0
            )

            # if length of iqr is close to zero, do not scale this feature
            self.scale_ = self.iqr_
            tol = _tol_wrt_dtype(X)
            if ht.abs(self.scale_).min() < tol:
                self.scale_ = ht.where(
                    ht.abs(self.scale_) >= tol, self.scale_, ht.ones_like(self.scale_)
                )
                print(
                    "At least one of the features is almost constant (w.r.t. machine precision) and will not be scaled for this reason."
                )
            self.scale_ = 1.0 / self.scale_
        return self

    def transform(self, X: ht.DNDarray) -> Union[Self, ht.DNDarray]:
        """
        Transform given data with RobustScaler

        Parameters
        ----------
        X : DNDarray of shape (n_datapoints, n_features)
            Data set to be transformed.
        """
        _is_2D_float_DNDarray(X)
        if self.with_centering:
            _has_n_features(self.center_, X)
        if self.with_scaling:
            _has_n_features(self.scale_, X)
        if self.copy:
            Y = X.copy()
            if self.with_centering:
                Y -= self.center_
            if self.with_scaling:
                Y *= self.scale_
            return Y
        # else in-place:
        if self.with_centering:
            X -= self.center_
        if self.with_scaling:
            X *= self.scale_
        return X

    def inverse_transform(self, Y: ht.DNDarray) -> Union[Self, ht.DNDarray]:
        """
        Apply inverse of :meth:``transform``.

        Parameters
        ----------
        Y : DNDarray of shape (n_datapoints, n_features)
            Data to be back-transformed
        """
        _is_2D_float_DNDarray(Y)
        if self.with_centering:
            _has_n_features(self.center_, Y)
        if self.with_scaling:
            _has_n_features(self.scale_, Y)
        if self.copy:
            X = Y.copy()
            if self.with_scaling:
                X /= self.scale_
            if self.with_centering:
                X += self.center_
            return X
        # else in-place:
        if self.with_scaling:
            Y /= self.scale_
        if self.with_centering:
            Y += self.center_
        return Y



"""Create a sperical dataset."""

import heat as ht
import torch


def create_spherical_dataset(
    num_samples_cluster, radius=1.0, offset=4.0, dtype=ht.float32, random_state=1
):
    """
    Creates k=4 sperical clusters in 3D space along the space-diagonal

    Parameters
    ----------
    num_samples_cluster: int
        Number of samples per cluster. Each process will create n // MPI_WORLD.size elements for each cluster
    radius: float
        Radius of the sphere
    offset: float
        Shift of the clusters along the axes. The 4 clusters will be positioned centered around c1=(offset, offset,offset),
        c2=(2*offset,2*offset,2*offset), c3=(-offset, -offset, -offset) and c4=(2*offset, -2*offset, -2*offset)
    dtype: ht.datatype
        Dataset dtype
    random_state: int
        seed of the torch random number generator
    """
    # contains num_samples

    p = ht.MPI_WORLD.size
    # create k sperical clusters with each n elements per cluster. Each process creates k * n/p elements
    num_ele = num_samples_cluster // p
    ht.random.seed(random_state)
    # radius between 0 and 1
    r = ht.random.rand(num_ele, split=0) * radius
    # theta between 0 and pi
    theta = ht.random.rand(num_ele, split=0) * 3.1415
    # phi between 0 and 2pi
    phi = ht.random.rand(num_ele, split=0) * 2 * 3.1415
    # Cartesian coordinates
    x = r * ht.sin(theta) * ht.cos(phi)
    x.astype(dtype, copy=False)
    y = r * ht.sin(theta) * ht.sin(phi)
    y.astype(dtype, copy=False)
    z = r * ht.cos(theta)
    z.astype(dtype, copy=False)

    cluster1 = ht.stack((x + offset, y + offset, z + offset), axis=1)
    cluster2 = ht.stack((x + 2 * offset, y + 2 * offset, z + 2 * offset), axis=1)
    cluster3 = ht.stack((x - offset, y - offset, z - offset), axis=1)
    cluster4 = ht.stack((x - 2 * offset, y - 2 * offset, z - 2 * offset), axis=1)

    data = ht.concatenate((cluster1, cluster2, cluster3, cluster4), axis=0)
    # Note: enhance when shuffel is available
    return data


def create_clusters(
    n_samples, n_features, n_clusters, cluster_mean, cluster_std, cluster_weight=None, device=None
):
    """
    Creates a DNDarray of shape (n_samples, n_features), split=0, and dtype=ht.float32, that is balanced (i.e. roughly same size of samples on each process).
    The data set consists of n_clusters clusters, each of which is sampled from a multivariate normal distribution with mean cluster_mean[k,:] and covariance matrix cluster_std[k,:,:].
    The clusters are of the same size (quantitatively) and distributed evenly over the processes, unless cluster_weight is specified.

    Parameters
    ------------
    n_samples: int
        Number of overall samples
    n_features: int
        Number of features
    n_clusters: int
        Number of clusters
    cluster_mean: torch.Tensor of shape (n_clusters, n_features)
        featurewise mean (center) of each cluster; of course not the true mean, but rather the mean according to which the elements of the cluster are sampled.
    cluster_std: torch.Tensor of shape (n_clusters, n_features, n_features), or (n_clusters,)
        featurewise standard deviation of each cluster from the mean value; of course not the true std, but rather the std according to which the elements of the cluster are sampled.
        If shape is (n_clusters,), std is assumed to be the same in each direction for each cluster
    cluster_weight: torch.Tensor of shape (n_clusters,), optional
        On each process, cluster_weight is assumed to be a torch.Tensor whose entries add up to 1. The i-th entry of cluster_weight on process p specified which amount of the samples on process p
        is sampled according to the distribution of cluster i. Thus, this parameter allows to distribute the n_cluster clusters unevenly over the processes.
        If None, each cluster is distributed evenly over all processes.
    device: Optional[str] = None,
        The device on which the data is stored. If None, the default device is used.
    """
    device = ht.devices.sanitize_device(device)

    if cluster_weight is None:
        cluster_weight = torch.ones(n_clusters) / n_clusters
    else:
        if not isinstance(cluster_weight, torch.Tensor):
            raise TypeError(
                "cluster_weight must be None or a torch.Tensor, but is {}".format(
                    type(cluster_weight)
                )
            )
        elif not cluster_weight.shape == (n_clusters,):
            raise ValueError(
                "If a torch.Tensor, cluster_weight must be of shape (n_clusters,), but is {}".format(
                    cluster_weight.shape
                )
            )
        elif not torch.allclose(torch.sum(cluster_weight), torch.tensor(1.0)):
            raise ValueError(
                "If a torch.Tensor, cluster_weight must add up to 1, but adds up to {}".format(
                    torch.sum(cluster_weight)
                )
            )
    if not isinstance(cluster_mean, torch.Tensor):
        raise TypeError("cluster_mean must be a torch.Tensor, but is {}".format(type(cluster_mean)))
    elif not cluster_mean.shape == (n_clusters, n_features):
        raise ValueError(
            "cluster_mean must be of shape (n_clusters, n_features), but is {}".format(
                cluster_mean.shape
            )
        )
    if not isinstance(cluster_std, torch.Tensor):
        raise TypeError("cluster_std must be a torch.Tensor, but is {}".format(type(cluster_std)))
    elif not cluster_std.shape == (
        n_clusters,
        n_features,
        n_features,
    ) and not cluster_std.shape == (n_clusters,):
        raise ValueError(
            "cluster_std must be of shape (n_clusters, n_features, n_features) or (n_clusters,), but is {}".format(
                cluster_std.shape
            )
        )
    if cluster_std.shape == (n_clusters,):
        cluster_std = torch.stack(
            [torch.eye(n_features) * cluster_std[k] for k in range(n_clusters)], dim=0
        )

    global_shape = (n_samples, n_features)
    local_shape = ht.MPI_WORLD.chunk(global_shape, 0)[1]
    local_size_of_clusters = [int(local_shape[0] * cluster_weight[k]) for k in range(n_clusters)]
    if sum(local_size_of_clusters) != local_shape[0]:
        local_size_of_clusters[0] += local_shape[0] - sum(local_size_of_clusters)
    distributions = [
        torch.distributions.multivariate_normal.MultivariateNormal(
            cluster_mean[k, :], cluster_std[k]
        )
        for k in range(n_clusters)
    ]
    local_data = [
        distributions[k].sample((local_size_of_clusters[k],)).to(device.torch_device)
        for k in range(n_clusters)
    ]
    local_data = torch.cat(local_data, dim=0)
    rand_perm = torch.randperm(local_shape[0])
    local_data = local_data[rand_perm, :]
    data = ht.DNDarray(
        local_data,
        global_shape,
        dtype=ht.float32,
        split=0,
        device=device,
        comm=ht.MPI_WORLD,
        balanced=True,
    )
    return data
